// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Nathan van Beelen
// Copyright (C) 2019  Mathieu Velten (FLA signed)
//
// This file is part of matrix-dart-sdk-sqflite.
//
// matrix-dart-sdk-sqflite is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk-sqflite is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk-sqflite.  If not, see <https://www.gnu.org/licenses/>.

library matrix_sdk_sqflite;

import 'dart:convert';

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:meta/meta.dart';
import 'package:sqflite/sqflite.dart';
import 'package:url/url.dart';

class SqfliteStore implements Store {

  final String path;
  SqfliteStore({@required this.path});

  bool _isOpen = false;

  @override
  bool get isOpen => _isOpen;

  LocalUser _user;

  Database _db;

  static const sent = 'sent';
  static const unsent = 'unsent';

  static const _userInfoTable = 'user_info';
  static const _roomTable = 'room';
  static const _eventTable = 'event';
  static const _userTable = 'user';
  static const _memberTable = 'member';
  static const _userStateTable = 'user_state';
  static const _customDataTable = 'custom_data';

  static const _currentUserStateView = 'current_user_state';

  static const _idColumn = 'id';

  static const _userIdColumn = 'user_id';
  static const _accessTokenColumn = 'access_token';
  static const _syncTokenColumn = 'sync_token';
  static const _homeserverColumn = 'homeserver';
  static const _avatarUrlColumn = 'avatar_url';

  static const _roomIdColumn = 'room_id';

  // Room columns
  static const _typeColumn = 'type';
  static const _nameColumn = 'name';
  static const _directUserIdColumn = 'direct_user_id';
  static const _previousBatchColumn = 'previous_batch';
  static const _creatorIdColumn = 'creator_id';
  static const _creatorStateSinceColumn = 'creator_state_since';
  static const _creatorNameColumn = 'creator_name';
  static const _creatorAvatarUrlColumn = 'creator_avatar_url';

  // Event columns
  static const _contentColumn = 'content';
  static const _senderIdColumn = 'sender_id';
  static const _timeColumn = 'time';
  static const _sentStateColumn = 'sent_state';
  static const _senderStateSinceColumn = 'sender_state_since';

  // UserState columns
  static const _sinceColumn = 'since';
  static const _displayNameColumn = 'display_name';

  // Custom data columns
  static const _keyColumn = 'key';
  static const _dataColumn = 'data';
  static const _objectIdColumn = 'object_id';

  Future<void> _createDatabase(Database db, LocalUser user) async {
    if (user == null) {
      throw ArgumentError(
        'user cannot be null if the database does not exist'
      );
    }

    _user = user;

    final batch = db.batch()
    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_userInfoTable (
        $_idColumn INTEGER PRIMARY KEY,
        $_userIdColumn TEXT,
        $_accessTokenColumn TEXT,
        $_syncTokenColumn TEXT,
        $_homeserverColumn TEXT);
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_roomTable (
          $_idColumn TEXT PRIMARY KEY,
          $_typeColumn TEXT,
          $_nameColumn TEXT NULL,
          $_avatarUrlColumn TEXT NULL,
          $_directUserIdColumn TEXT NULL,
          $_creatorIdColumn TEXT,
          $_creatorStateSinceColumn INT NULL,
          $_previousBatchColumn TEXT NOT NULL,
          FOREIGN KEY ($_directUserIdColumn) REFERENCES $_userTable($_idColumn));
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_eventTable (
          $_idColumn TEXT PRIMARY KEY,
          $_roomIdColumn TEXT,
          $_senderIdColumn TEXT NULL,
          $_senderStateSinceColumn TEXT NOT NULL,
          $_timeColumn INT NOT NULL,
          $_contentColumn TEXT NOT NULL,
          $_sentStateColumn TEXT NOT NULL,
          FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn),
          FOREIGN KEY($_senderIdColumn, $_roomIdColumn, $_senderStateSinceColumn)
            REFERENCES $_userStateTable($_userIdColumn,
                                        $_roomIdColumn,
                                        $_sinceColumn)
        );
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_userTable (
          $_idColumn TEXT PRIMARY KEY);
      ''')

    ..execute(
      '''
      CREATE TABLE IF NOT EXISTS $_userStateTable (
        $_userIdColumn TEXT,
        $_roomIdColumn TEXT,
        $_sinceColumn INT,
        $_displayNameColumn TEXT NULL,
        $_avatarUrlColumn TEXT NULL,
        FOREIGN KEY($_userIdColumn) REFERENCES $_userTable($_idColumn),
        FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn)
        PRIMARY KEY ($_userIdColumn, $_roomIdColumn, $_sinceColumn));
      ''')

    ..execute(
      '''
      CREATE VIEW IF NOT EXISTS $_currentUserStateView AS
      SELECT $_userIdColumn,
             $_roomIdColumn,
             $_displayNameColumn,
             $_avatarUrlColumn,
             MAX($_sinceColumn) AS $_sinceColumn
      FROM $_userStateTable
      GROUP BY $_userIdColumn, $_roomIdColumn
      '''
    )

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_memberTable (
          $_userIdColumn TEXT,
          $_roomIdColumn TEXT,
          FOREIGN KEY($_userIdColumn) REFERENCES $_userTable($_idColumn),
          FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn)
          PRIMARY KEY ($_userIdColumn, $_roomIdColumn));
      ''')

    ..execute(
      '''
        CREATE TABLE IF NOT EXISTS $_customDataTable (
          $_objectIdColumn TEXT NULL,
          $_keyColumn,
          $_dataColumn TEXT NULL,
          PRIMARY KEY ($_objectIdColumn, $_keyColumn));
          );
      ''');

    await batch.commit(noResult: true);
  }

  Future<LocalUser> _userFromDb(Database db) async {
    final userId = await getUserId(db: db);
    final hs = await getHomeserver(db: db);
    final accessToken = await getAccessToken(db: db);

    return LocalUser(
        id: userId,
        homeserver: hs,
        accessToken: accessToken,
        store: this
    );
  }

  @override
  Future<bool> open({LocalUser user}) async {
    try {
      _db = await openDatabase(
        path,
        version: 12,
        onUpgrade: (Database db, int oldVersion, int newVersion) async {

          // First get the user from the old database
          final localUser = await _userFromDb(db);

          // Drop all tables and delete the sync token
          final batch = db.batch()
            ..execute('UPDATE $_userInfoTable SET $_syncTokenColumn=NULL')
            ..execute('DROP TABLE IF EXISTS $_roomTable')
            ..execute('DROP TABLE IF EXISTS $_eventTable')
            ..execute('DROP TABLE IF EXISTS $_userTable')
            ..execute('DROP TABLE IF EXISTS $_memberTable')
            ..execute('DROP TABLE IF EXISTS $_userStateTable')
            ..execute('DROP VIEW  IF EXISTS $_currentUserStateView');

          await batch.commit(noResult: true);

          await _createDatabase(db, localUser);
        },
        onCreate: (Database db, int version) async {
          await _createDatabase(db, user);
        }
      );

      if (_user == null) {
        _user = await _userFromDb(_db);
      }
    } on ArgumentError {
      return false;
    }

    _isOpen = true;
    return true;
  }

  @override
  Future<void> close() async {
    _isOpen = false;
    _db.close();
  }

  @override
  Future<void> delete() async {
    await deleteDatabase(await getDatabasesPath());
  }

  Future<String> _getUserInfo(String column, {Database db}) async {
    if (db == null) {
      db = _db;
    }

    var results = await db.rawQuery(
      '''
      SELECT $column FROM $_userInfoTable WHERE $_idColumn = 1;
      '''
    );

    if (results.isNotEmpty) {
      return results.first[column];
    } else {
      return null;
    }
  }

  Future<void> _setUserInfo(String column, String value, {Database db}) async {
    if (db == null) {
      db = _db;
    }

    final batch = db.batch()
    ..execute(
      '''
      UPDATE $_userInfoTable SET $column = ?;
      ''',
      [value]
    )

    ..execute(
      '''
      INSERT OR IGNORE INTO $_userInfoTable($_idColumn, $column)
        values (?, ?);
      ''',
      [1, value]
    );

    await batch.commit(noResult: true);
  }

  @override
  Future<UserId> getUserId({Database db}) async {
    final value = await _getUserInfo(_userIdColumn, db: db);
    if (value == null) {
      return null;
    }

    return UserId(value);
  }

  @override
  Future<void> setUserId(UserId userId) async {
    await _setUserInfo(_userIdColumn, userId.toString());
  }

  @override
  Future<String> getAccessToken({Database db}) async {
    return _getUserInfo(_accessTokenColumn, db: db);
  }

  @override
  Future<void> setAccessToken(String accessToken) async {
    await _setUserInfo(_accessTokenColumn, accessToken);
  }

  @override
  Future<String> getSyncToken() async {
    return _getUserInfo(_syncTokenColumn);
  }

  @override
  Future<void> setSyncToken(String syncToken) async {
    await _setUserInfo(_syncTokenColumn, syncToken);
  }

  @override
  Future<Homeserver> getHomeserver({Database db}) async {
    final value = await _getUserInfo(_homeserverColumn, db: db);
    if (value == null) {
      return null;
    }

    return Homeserver(Url.parse(value));
  }

  @override
  Future<void> setHomeserver(Homeserver homeserver) async {
    await _setUserInfo(_homeserverColumn, homeserver.url.toString());
  }

  static const _userAvatarUrlColumn = 'user_avatar_url';

  static const _roomSelectBase =
    '''
    SELECT $_roomTable.$_idColumn,
           $_roomTable.$_typeColumn,
           $_roomTable.$_nameColumn,
           $_roomTable.$_avatarUrlColumn,
           $_roomTable.$_directUserIdColumn,
           $_roomTable.$_previousBatchColumn,
           $_roomTable.$_creatorIdColumn,
           $_roomTable.$_creatorStateSinceColumn,
           $_currentUserStateView.$_displayNameColumn,
           $_currentUserStateView.$_avatarUrlColumn AS $_userAvatarUrlColumn,
           $_currentUserStateView.$_sinceColumn,
           $_userStateTable.$_displayNameColumn AS $_creatorNameColumn,
           $_userStateTable.$_avatarUrlColumn AS $_creatorAvatarUrlColumn
    FROM $_roomTable
    LEFT JOIN $_currentUserStateView
      ON $_currentUserStateView.$_userIdColumn=$_roomTable.$_directUserIdColumn
       AND $_currentUserStateView.$_roomIdColumn=$_roomTable.$_idColumn
    LEFT JOIN $_userStateTable
      ON $_userStateTable.$_userIdColumn=$_roomTable.$_creatorIdColumn
       AND $_userStateTable.$_roomIdColumn=$_roomTable.$_idColumn
       AND $_userStateTable.$_sinceColumn=$_roomTable.$_creatorStateSinceColumn
    ''';

  Room _parseRoomRow(Map<String, dynamic> row) {
    final id = RoomId(row[_idColumn]);
    final type = row[_typeColumn];

    final name = row[_nameColumn];

    var avatarUrl = row[_avatarUrlColumn];
    if (avatarUrl != null) {
      avatarUrl = Url.parse(avatarUrl);
    }

    final previousBatch = row[_previousBatchColumn];

    final directUserId = row[_directUserIdColumn];
    User directUser;
    if (directUserId != null) {
      final displayName = row[_displayNameColumn];

      var userAvatarUrl = row[_userAvatarUrlColumn];
      if (userAvatarUrl != null) {
        userAvatarUrl = Url.parse(userAvatarUrl);
      }

      var since = row[_sinceColumn];
      if (since != null) {
        since = DateTime.fromMillisecondsSinceEpoch(since);
      }

      directUser = User(
        id: UserId(directUserId),
        state: UserState(
          roomId: id,
          since: since,
          displayName: displayName,
          avatarUrl: userAvatarUrl
        )
      );
    }

    final creatorId = UserId(row[_creatorIdColumn]);
    final creatorName = row[_creatorNameColumn];
    var creatorAvatarUrl = row[_creatorAvatarUrlColumn];
    if (creatorAvatarUrl != null) {
      creatorAvatarUrl = Url.parse(creatorAvatarUrl);
    }

    final creatorStateSince
      = DateTime.fromMillisecondsSinceEpoch(row[_creatorStateSinceColumn]);

    final creator = User(
      id: creatorId,
      state: UserState(
        roomId: id,
        since: creatorStateSince,
        displayName: creatorName,
        avatarUrl: creatorAvatarUrl
      )
    );

    if (type == 'join') {
      return JoinedRoom(
        id: id,
        name: name,
        avatarUrl: avatarUrl,
        previousBatch: previousBatch,
        scope: _user.scope,
        directUser: directUser,
        creator: creator
      );
    } else {
      return LeftRoom(
        id: id,
        name: name,
        avatarUrl: avatarUrl,
        previousBatch: previousBatch,
        scope: _user.scope,
        directUser: directUser,
        creator: creator
      );
    }

  }

  @override
  Future<Room> getRoom(RoomId id) async {
    var results = await _db.rawQuery(
      _roomSelectBase + ' WHERE $_roomTable.$_idColumn = ?;',
      [id.toString()]
    );

    if (results.isNotEmpty) {
      return _parseRoomRow(results.first);
    } else {
      return null;
    }
  }

  @override
  Stream<Room> getRooms() async* {
    var results = await _db.rawQuery(_roomSelectBase + ';');

    if (results.isNotEmpty) {
      for (var result in results) {
        yield _parseRoomRow(result);
      }
    }
  }

  @override
  Future<void> setRooms(Iterable<Room> rooms) async {
    await _db.transaction((txn) async {
      for (Room room in rooms) {
        var url;
        if (room.avatarUrl != null) {
          url = room.avatarUrl.toString();
        }

        var type;
        if (room is JoinedRoom) {
          type = 'join';
        } else if (room is LeftRoom) {
          type = 'leave';
        }

        await txn.execute(
          '''
          UPDATE $_roomTable SET $_typeColumn = ?,
                                 $_nameColumn = ?,
                                 $_avatarUrlColumn = ?,
                                 $_directUserIdColumn = ?,
                                 $_creatorIdColumn = ?,
                                 $_creatorStateSinceColumn = ?,
                                 $_previousBatchColumn = ?
                             WHERE $_idColumn = ?;
          ''',
          [
            type,
            room.name,
            url,
            room.directUser?.id?.toString(),
            room.creator.id.toString(),
            room.creator.state.since.millisecondsSinceEpoch,
            room.previousBatch,
            room.id.toString()
          ]
        );

        await txn.execute(
          '''
          INSERT OR IGNORE INTO $_roomTable($_idColumn,
                                            $_typeColumn,
                                            $_nameColumn,
                                            $_avatarUrlColumn,
                                            $_directUserIdColumn,
                                            $_creatorIdColumn,
                                            $_creatorStateSinceColumn,
                                            $_previousBatchColumn)
            values (?, ?, ?, ?, ?, ?, ?, ?);
          ''',
          [
            room.id.toString(),
            type,
            room.name,
            url,
            room.directUser?.id?.toString(),
            room.creator.id.toString(),
            room.creator.state.since.millisecondsSinceEpoch,
            room.previousBatch
          ]
        );
      }
    });
  }

  @override
  Future<void> removeEvent(EventId id) async {
    await _db.rawQuery(
      '''
      DELETE FROM $_eventTable WHERE $_idColumn = ?;
      ''',
      [id.toString()]
    );
  }

  @override
  Future<RoomEvent> getRoomEvent(Room room, EventId id) async {
    var results = await _db.rawQuery(
      '''
      SELECT $_contentColumn FROM $_eventTable WHERE $_idColumn = ?
                                               AND $_roomIdColumn = ?;
      ''',
      [id.toString(), room.id.toString()]
    );

    if (results.isNotEmpty) {
      var content = results.first[_contentColumn];
      return RoomEvent.fromJson(RoomEventArgs(room: room), json.decode(content));
    } else {
      return null;
    }
  }

  @override
  Stream<RoomEvent> getRoomEvents(Room room, {int count, RoomEvent until}) async* {
    final context = ProcessContext();

    var query =
    '''
      SELECT $_eventTable.$_contentColumn,
             $_eventTable.$_sentStateColumn,
             $_eventTable.$_senderIdColumn,
             $_userStateTable.$_sinceColumn,
             $_userStateTable.$_displayNameColumn,
             $_userStateTable.$_avatarUrlColumn
      FROM $_eventTable
      LEFT OUTER JOIN $_userStateTable
        ON $_userStateTable.$_userIdColumn=$_eventTable.$_senderIdColumn
          AND $_userStateTable.$_roomIdColumn=$_eventTable.$_roomIdColumn
          AND $_userStateTable.$_sinceColumn=$_eventTable.$_senderStateSinceColumn
      WHERE $_eventTable.$_roomIdColumn = ?
    ''';

    var values = <dynamic>[room.id.toString()];

    if (until != null) {
      query += ' AND $_eventTable.$_timeColumn >= ?';
      values.add(until.time.millisecondsSinceEpoch);
    }

    query += ' ORDER BY $_eventTable.$_timeColumn DESC';

    if (count != null) {
      query += ' LIMIT ?';
      values.add(count);
    }

    var results = await _db.rawQuery(query, values);

    if (results.isNotEmpty) {
      for (final result in results) {
        final content = result[_contentColumn];
        final sentState = result[_sentStateColumn] == 'sent'
                          ? SentState.sent : SentState.unsent;

        final displayName = result[_displayNameColumn];
        var avatarUrl = result[_avatarUrlColumn];
        if (avatarUrl != null) {
          avatarUrl = Url.parse(avatarUrl);
        }

        var since = result[_sinceColumn];
        if (since != null) {
          since = DateTime.fromMillisecondsSinceEpoch(since);
        }

        final user = User(
          id: UserId(result[_senderIdColumn]),
          state: UserState(
            roomId: room.id,
            since: since,
            displayName: displayName,
            avatarUrl: avatarUrl
          )
        );

        yield await RoomEvent.fromJson(
          RoomEventArgs(
            room: room,
            sender: user,
            sentState: sentState
          ),
          json.decode(content),
          context: context
        );
      }
    }
  }

  @override
  Stream<RoomEvent> getUnsentEvents(Room room) async* {
    final results = await _db.rawQuery(
    '''
      SELECT $_eventTable.$_contentColumn
      FROM $_eventTable
      WHERE $_eventTable.$_roomIdColumn = ?
        AND $_eventTable.$_sentStateColumn = ?
    ''',
    [room.id.toString(), unsent]);

    if (results.isNotEmpty) {
      for (final result in results) {
        final content = result[_contentColumn];

        yield await RoomEvent.fromJson(RoomEventArgs(
          room: room,
          sender: _user,
          sentState: SentState.unsent
        ), json.decode(content));
      }
    }
  }

  @override
  Future<void> addRoomEvents(Room room, Iterable<RoomEvent> events) async {
    await _db.transaction((txn) {
      for (Event event in events) {
        var senderId;
        var senderStateSince;

        var time;
        final sentState = event.sentState != SentState.sent ? unsent : sent;

        if (event is RoomEvent) {
          senderId = event.sender.id.toString();
          time = event.time.millisecondsSinceEpoch;
          senderStateSince = event.sender.state?.since?.millisecondsSinceEpoch;
        }

        txn.execute(
          '''
          INSERT OR IGNORE INTO $_eventTable($_idColumn,
                                             $_roomIdColumn,
                                             $_timeColumn,
                                             $_senderIdColumn,
                                             $_senderStateSinceColumn,
                                             $_contentColumn,
                                             $_sentStateColumn)
            values (?, ?, ?, ?, ?, ?, ?);
          ''',
          [
            event.id.toString(),
            room.id.toString(),
            time,
            senderId,
            senderStateSince,
            json.encode(event.toJson()),
            sentState,
          ]
        );
      }
    });
  }

  Future<void> _setUser(Transaction txn, User user) async {
    var url;
    if (user.avatarUrl != null) {
      url = user.avatarUrl.toString();
    }

    txn.execute(
      '''
      INSERT OR IGNORE INTO $_userTable($_idColumn)
        VALUES (?);
      ''',
      [user.id.toString()]
    );

    if (user.state != null) {
      txn.execute(
      '''
      INSERT OR IGNORE INTO $_userStateTable($_userIdColumn,
                                             $_roomIdColumn,
                                             $_sinceColumn,
                                             $_displayNameColumn,
                                             $_avatarUrlColumn)
        VALUES (?, ?, ?, ?, ?);
      ''',
          [
            user.id.toString(),
            user.state.roomId.toString(),
            user.state.since.millisecondsSinceEpoch,
            user.state.displayName,
            url
          ]
      );
    }
  }

  static const _memberSelectBase =
    '''
    SELECT $_memberTable.$_userIdColumn,
           $_currentUserStateView.$_roomIdColumn,
           $_currentUserStateView.$_displayNameColumn,
           $_currentUserStateView.$_avatarUrlColumn,
           $_currentUserStateView.$_sinceColumn
    FROM $_memberTable
      LEFT JOIN $_currentUserStateView
        ON
         $_currentUserStateView.$_userIdColumn=$_memberTable.$_userIdColumn
          AND
           $_currentUserStateView.$_roomIdColumn=$_memberTable.$_roomIdColumn
    WHERE $_memberTable.$_roomIdColumn = ?
    ''';

  @override
  Future<User> getMember(RoomId roomId, UserId id, {DateTime at}) async {
    var query =
      _memberSelectBase +
       '''
       AND $_memberTable.$_userIdColumn = ?
       ''';

    var values = [roomId.toString(), id.toString()];

    final results = await _db.rawQuery(query, values);

    if (results.isNotEmpty) {
      final first = results.first;
      final user = _parseUserRow(first, id);

      return user;
    } else {
      return null;
    }
  }

  @override
  Stream<User> getMembers(RoomId roomId) async* {
    final results = await _db.rawQuery(
      _memberSelectBase,
      [roomId.toString()]
    );

    if (results.isNotEmpty) {
      for (final result in results) {
        final id = UserId(result[_userIdColumn]);
        yield _parseUserRow(result, id);
      }
    }

  }

  User _parseUserRow(Map<String, dynamic> row, UserId id) {

    var roomId = row[_roomIdColumn];
    if (roomId != null) {
      roomId = RoomId(roomId);
    }

    var since = row[_sinceColumn];
    if (since != null) {
      since = DateTime.fromMicrosecondsSinceEpoch(row[_sinceColumn]);
    }

    final displayName = row[_displayNameColumn];
    var avatarUrl = row[_avatarUrlColumn];
    if (avatarUrl != null) {
      avatarUrl = Url.parse(avatarUrl);
    }

    return User(
      id: id,
      state: UserState(
        roomId: roomId,
        since: since,
        displayName: displayName,
        avatarUrl: avatarUrl
      )
    );
  }


  @override
  Future<void> addMembers(RoomId roomId, Iterable<User> members) async {
    await _db.transaction((txn) {
      for (User member in members) {
        _setUser(txn, member);

        txn.execute(
          '''
          INSERT OR IGNORE INTO $_memberTable($_userIdColumn, 
                                              $_roomIdColumn)
            values (?, ?);
          ''',
          [member.id.toString(), roomId.toString()]
        );
      }
    });
  }

  @override
  Future<String> getObjectCustomData(Id objectId, String key) async {
    if (objectId == null) {
      throw Exception();
    }

    var results = await _db.rawQuery(
      '''
      SELECT * FROM $_customDataTable
      WHERE $_customDataTable.$_objectIdColumn = ?
      AND $_customDataTable.$_keyColumn = ?;
      ''',
      [objectId.toString(), key]
    );

    if (results.isNotEmpty) {
      return results[0][_dataColumn];
    }
    return null;
  }

  @override
  Future<void> setObjectCustomData(Id objectId, String key, String value) async {
    if (objectId == null) {
      throw Exception();
    }
    await _db.transaction((txn) {
        txn.execute(
          '''
          INSERT OR REPLACE INTO $_customDataTable($_objectIdColumn,
                                                 $_keyColumn,
                                                 $_dataColumn)
            values (?, ?, ?);
          ''',
          [objectId.toString(), key, value]
        );
    });
  }

  @override
  Future<String> getAppCustomData(String key) async {
    var results = await _db.rawQuery(
      '''
      SELECT * FROM $_customDataTable
      WHERE $_customDataTable.$_keyColumn = ?
      ''',
      [key]
    );

    if (results.isNotEmpty) {
      return results[0][_dataColumn];
    }
    return null;
  }

  @override
  Future<void> setAppCustomData(String key, String value) async {
    await _db.transaction((txn) {
        txn.execute(
          '''
          INSERT OR REPLACE INTO $_customDataTable($_keyColumn,
                                                $_dataColumn)
            values (?, ?);
          ''',
          [key, value]
        );
    });
  }
}
